////////////////////////////////////////////////////////////////////////////////
//
//  Copyright 2013 by Michael A. Morris, dba M. A. Morris & Associates
//
//  All rights reserved. The source code contained herein is publicly released
//  under the terms and conditions of the GNU Lesser Public License. No part of
//  this source code may be reproduced or transmitted in any form or by any
//  means, electronic or mechanical, including photocopying, recording, or any
//  information storage and retrieval system in violation of the license under
//  which the source code is released.
//
//  The source code contained herein is free; it may be redistributed and/or
//  modified in accordance with the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either version 2.1 of
//  the GNU Lesser General Public License, or any later version.
//
//  The source code contained herein is freely released WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
//  PARTICULAR PURPOSE. (Refer to the GNU Lesser General Public License for
//  more details.)
//
//  A copy of the GNU Lesser General Public License should have been received
//  along with the source code contained herein; if not, a copy can be obtained
//  by writing to:
//
//  Free Software Foundation, Inc.
//  51 Franklin Street, Fifth Floor
//  Boston, MA  02110-1301 USA
//
//  Further, no use of this source code is permitted in any form or means
//  without inclusion of this banner prominently in any derived works.
//
//  Michael A. Morris
//  Huntsville, AL
//
////////////////////////////////////////////////////////////////////////////////

`timescale 1ns / 1ps

////////////////////////////////////////////////////////////////////////////////
// Company:         Michael A. Morris 
// Engineer:        M. A. Morris & Associates 
// 
// Create Date:     15:28:27 09/28/2013 
// Design Name:     ChoChi FPGA-based 6502 Development Board
// Module Name:     ChoChi_MemoryIF.v 
// Project Name:    C:\XProjects\ISE10.1i\M65C02A 
// Target Devices:  SRAM-based FPGAs: XC3S50-5VQ100; XC3S50A-4VQ100 
// Tool versions:   Xilinx ISE 10.1i
//    
// Description:
//
//  This module provides an interface to an external asynchronous SRAM which
//  emulates the behavior of internal Block RAM.
//
//  For memory read cycles, the expectation is that the address is provided
//  after the rising edge of Clk during the first clock period, and the input
//  data (from the SRAM) is available after the next rising edge of Clk.
//
//  For memory write cycles, the expectation is that the address and the output
//  data (to the SRAM) is provided after the rising edge of Clk during one clock
//  period. The output data is written to the external RAM during that period,
//  and the memory interface is ready for the next memory operation at the next
//  rising edge of Clk.
//
//  It is assumed that an asynchronous multiplexer supplies the input address
//  and output data. It is further assumed that the input data is connected to a
//  wired-OR bus. Finally, it is assumed that the objective is single cycle, no
//  wait state memory cycles.
//
//  The first assumption requires that the memory address and output data be
//  registered on the falling edge of Clk in order to provide the external SRAM 
//  stable addresses for at least half of the clock period. It then follows that
//  the external SRAM's chip enable (nRAM_CE), output enable (nRAM_OE), and
//  write enable (nRAM_WE) control signals be driven between the falling and
//  rising edges of the system clock. This allows the external SRAM read/write
//  cycles to be controlled by the falling edges of chip enable/output enable 
//  for memory read cycles, and the rising edge of chip enable/write enable for
//  memory write cycles.
//
//  The second assumption requires that the input data register be reset/cleared
//  whenever the external memory interface is not selected: (Sel == 0). A clear,
//  or asynchronous active low reset, should be used to provide the best margins
//  on the wired-OR bus to which the output of input data register is connected.
//
//  The third assumption means that the module will not generate an acknowledge,
//  or other kind of ready signal, to control a wait state generator which is
//  external to the module. Thus, the period of the clock will have to be adjus-
//  ted to accomodate the timing requirements of the external SRAM selected for
//  the application.
//
// Dependencies: 
//
// Revision:
//
//  0.01    13I28   MAM     File Created 
//
// Additional Comments: 
//
////////////////////////////////////////////////////////////////////////////////

module ChoChi_SRAM_IF #(
    parameter pMemAddr = 17
)(
    input   Rst,                        // System Reset
    input   Clk,                        // System Clk - 0 Delay
    input   Clk2x,                      // System Clk2x
    
    input   Sel,                        // SRAM IF Select       (Input)
    input   WE,                         // SRAM IF Write Enable (Input)
    input   RE,                         // SRAM IF Read Enable  (Input)
    input   [(pMemAddr - 1):0] A,       // SRAM IF Address      (Input)
    input   [7:0] DO,                   // SRAM IF Data Output  (Input)
    output  [7:0] DI,                   // SRAM IF Data Input   (Output)
    
    output  nRAM_CE,                    // SRAM Chip Enable     (Active Low Out)                    
    output  nRAM_OE,                    // SRAM Output Enable   (Active Low Out)
    output  nRAM_WE,                    // SRAM Write Enable    (Active Low Out)
    output  reg [(pMemAddr - 1):0] RAM_A,   // SRAM Address Bus (Output)
    inout   [7:0] RAM_DQ                // SRAM Data In/Out Bus (In/Out)
);

////////////////////////////////////////////////////////////////////////////////
//
//  Declarations
//

wire    RAM_OE;

wire    RAM_WE;
reg     RAM_DE;
reg     [7:0] RAM_DO_OFD;

reg     [7:0] RAM_DI;

////////////////////////////////////////////////////////////////////////////////
//
//  Implementation
//

//  Generate a pulse on the SRAM chip enable control signal using an ODDR2 
//      during the clock low period. ODDR2 function of the IOB is used to mini-
//      mize skew on the SRAM chip enable signal and to accurately control the
//      the pulse width. Since the output buffer delays can not be dynamically
//      adjusted, the ODDR2 provides a fixed pulse width and slew rate and
//      drive control parameters for the IOB must be adjusted to minimize the
//      fixed delays of the output signal path.

SRAM_StbGen CE_FF (
                .Rst(Rst),          // Set a high, inactive signal level on Rst
                .Clk0(~Clk),        // Set nRAM_CE 0 on FE(Clk) if Sel == 1 
                .Clk1( Clk),        // Set nRAM_CE 1 on RE(Clk) if Sel == 1
                .Stb(Sel),          // Controls the clock enable of ODDR2 FFs
                .nOut(nRAM_CE)      // Output of DDR2 multiplexer
            );
            
// Generate a pulse on the SRAM output enable control signals using an ODDR2
//      during the clock low period. (See comments above regarding skew, etc.) 
            
assign RAM_OE = Sel & RE;           // Qualify the SRAM output enable with Sel

SRAM_StbGen OE_FF (
                .Rst(Rst),          // Set a high, inactive signal level on Rst
                .Clk0(~Clk),        // Set nRAM_OE 0 on FE(Clk) if RAM_OE == 1 
                .Clk1( Clk),        // Set nRAM_OE 1 on RE(Clk) if RAM_OE == 1
                .Stb(RAM_OE),       // Controls the clock enable of ODDR2 FFs
                .nOut(nRAM_OE)      // Output of DDR2 multiplexer
            );

// Generate a pulse on the SRAM write enable control signals using an ODDR2
//      during the clock low period. (See comments above regarding skew, etc.) 
            
assign RAM_WE = Sel & WE;           // Qualify the SRAM write enable with Sel

SRAM_StbGen WE_FF (
                .Rst(Rst),          // Set a high, inactive signal level on Rst
                .Clk0(~Clk),        // Set nRAM_WE 0 on FE(Clk) if RAM_WE == 1
                .Clk1( Clk),        // Set nRAM_WE 1 on RE(Clk) if RAM_WE == 1
                .Stb(RAM_WE),       // Controls the clock enable of ODDR2 FFs
                .nOut(nRAM_WE)      // Output of DDR2 multiplexer
            );
            
//  Drive the external SRAM address signals
//      Capture the address on the falling edge of Clk
        
always @(negedge Clk)
begin
    if(Rst)
        RAM_A <= 0;
    else if(Sel)
        RAM_A <= A;
end

//  Capture the write data for the external SRAM
//      Capture the output data in an IOB FF on the falling edge of Clk

always @(negedge Clk)
begin
    if(Rst)
        RAM_DO_OFD <= 0;
    else if(RAM_WE)
        RAM_DO_OFD <= DO;
end

//  Drive the RAM_DQ output when RAM_DE is asserted, else tri-state

assign RAM_DE_In = (Clk & Sel & WE);

always @(posedge Clk2x or posedge Rst)
begin
    if(Rst)
        RAM_DE <= #1 0;
    else
        RAM_DE <= #1 RAM_DE_In;
end

assign RAM_DQ = ((RAM_DE) ? RAM_DO_OFD : 8'bZ);

//  Capture the external SRAM output data
//      Capture on rising edge of Clk in IOB input FF
//      Reset IOB FF whenever the interface is not selected
//          Simplifies the implementation of a wor bus in the FPGA

always @(posedge Clk or negedge Sel)
begin
    if(~Sel)
        RAM_DI <= #1 0;
    else
        RAM_DI <= #1 RAM_DQ;
end

assign DI = RAM_DI;

endmodule

