////////////////////////////////////////////////////////////////////////////////
//
//  Copyright 2013 by Michael A. Morris, dba M. A. Morris & Associates
//
//  All rights reserved. The source code contained herein is publicly released
//  under the terms and conditions of the GNU Lesser Public License. No part of
//  this source code may be reproduced or transmitted in any form or by any
//  means, electronic or mechanical, including photocopying, recording, or any
//  information storage and retrieval system in violation of the license under
//  which the source code is released.
//
//  The source code contained herein is free; it may be redistributed and/or
//  modified in accordance with the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either version 2.1 of
//  the GNU Lesser General Public License, or any later version.
//
//  The source code contained herein is freely released WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
//  PARTICULAR PURPOSE. (Refer to the GNU Lesser General Public License for
//  more details.)
//
//  A copy of the GNU Lesser General Public License should have been received
//  along with the source code contained herein; if not, a copy can be obtained
//  by writing to:
//
//  Free Software Foundation, Inc.
//  51 Franklin Street, Fifth Floor
//  Boston, MA  02110-1301 USA
//
//  Further, no use of this source code is permitted in any form or means
//  without inclusion of this banner prominently in any derived works.
//
//  Michael A. Morris
//  Huntsville, AL
//
////////////////////////////////////////////////////////////////////////////////

`timescale 1ns / 1ps

////////////////////////////////////////////////////////////////////////////////
// Company:         M. A. Morris & Associates 
// Engineer:        Michael A. Morris 
// 
// Create Date:     12:59:09 09/29/2013 
// Design Name:     ChoChi FPGA-based 6502 Development Board
// Module Name:     SRAM_IF_Tst.v 
// Project Name:    C:\XProjects\ISE10.1i\M65C02A 
// Target Devices:  SRAM-based FPGAs: XC3S50-5VQ100; XC3S50A-4VQ100 
// Tool versions:   Xilinx ISE 10.1i SP3
//    
// Description:
//
//  This module provides a synthesizable test bench for the ChoChi_SRAM_IF.v
//  module. It performs a walking 1s and walking 0s test on the SRAM data bus
//  and compares the value expected to be written with the value read. A regis-
//  ter is used to detect and hold any errors detected. An LED is driven low in
//  the event that an error is detected.
//
// Dependencies:    ClkGen.v
//                      ClkGen.xaw
//                      fedet.v
//                  ChoChi_SRAM_IF.v
//                      SRAM_StbGen.v
//                  SRAM_IF_Tst.ucf
//
// Revision: 
//
//  0.01    13I29   MAM     File Created
//
// Additional Comments:
//
//  Module also demonstrates how to use ODDR2 registers to generate Phi1O and 
//  Phi2O clocks from the basic internal clock, Clk. The resulting Phi1O and
//  Phi2O clocks have minimal skew, and a duty cycle that is very nearly 50%.
//  Also the overlap between the two phases is minimized using the ODDR2 tech-
//  nique.
//
//  ODDR2 registers are also used in the SRAM Interface module to provide non-
//  overlapping SRAM output and write enable strobes, and a synchronous SRAM
//  chip enable.
//
//  With appropriate gating of the RE and WE signals, the SRAM Interface module
//  can extend the bus cycle, i.e. insert wait states.
//
////////////////////////////////////////////////////////////////////////////////

module SRAM_IF_Tst #(
    parameter pMemAddr = 15
)(
    input   nRst,               // System Reset Input
    output  nRstO,              // Internal System Reset Output (OC w/ PU)
    input   ClkIn,              // System Clk Input

    output  Phi2O,              // Clock Phase 2 Output
    output  Phi1O,              // Clock Phase 1 Output - complement of Phi2O

    output  [3:0] nCE,          // Chip Enable for External RAM/ROM Memory
    output  RnW,                // Read/nWrite cycle control output signal
    output  nWr,                // External Asynchronous Bus Write Strobe
    output  nOE,                // External Asynchronous Bus
    output  [ 3:0] XA,          // Extended Address Output for External Memory
    output  [15:0] A,           // External Memory Address Bus
    inout   [ 7:0] DB,          // External, Bidirectional Data Bus

    output  [4:0] LED           // LED Test Register
);

////////////////////////////////////////////////////////////////////////////////
//
//  Declarations
//

wire    Rst;                    // System Reset
reg     OE_nRstO;               // Tri-state control for OC drive of nRstO
wire    Clk, Clk2x;             // System Clk, 2 x System Clk

reg     [19:0] Cntr;
reg     [ 7:0] DO, rDO;
wire    [ 7:0] DI;
reg     Err;

////////////////////////////////////////////////////////////////////////////////
//
//  Implementation
//

// Instantiate the Clk and Reset Generator Module

M65C02_ClkGen   ClkGen (
                    .nRst(nRst), 
                    .ClkIn(ClkIn),
                    
                    .Clk(Clk),              // Clk    <= (M/D) x ClkIn
                    .Clk90(),               // Quadrature Clk
                    .Clk2x(Clk2x),          // 2x Clk
                    .Buf_ClkIn(Buf_ClkIn),  // RefClk <= Buffered ClkIn

                    .Rst(Rst)
                );

//  Generate Reset output for use by external circuits

always @(posedge Buf_ClkIn or posedge Rst)
begin
    if(Rst)
        OE_nRstO <= #1 1;
    else
        OE_nRstO <= #1 0;
end

assign nRstO = ((OE_nRstO) ? 0 : 1'bZ);

// Instantiate the Unit Under Test (UUT) - ChoChi SRAM Interface

ChoChi_SRAM_IF  #(
                    .pMemAddr(pMemAddr)
                ) uut (
                    .Rst(Rst), 
                    .Clk(Clk),
                    .Clk2x(Clk2x),
                    
                    .Sel(Sel), 
                    .WE(WE), 
                    .RE(RE), 
                    .A(Cntr[19:5]), 
                    .DO(DO), 
                    .DI(DI), 

                    .nRAM_CE(nCE[0]), 
                    .nRAM_OE(nOE), 
                    .nRAM_WE(nWr), 
                    .RAM_A(A[14:0]), 
                    .RAM_DQ(DB)
                );
                
assign nCE[3:1] = 3'b111;
assign RnW      = Cntr[0];
assign XA[3:0]  = 4'b0000;

assign A[15:pMemAddr] = 1'b0;

always @(posedge Clk)
begin
    if(Rst)
        Cntr <= #1 0;
    else
        Cntr <= #1 (Cntr + 1);
end

assign Sel = ~Rst;
assign WE  = ~Cntr[0];
assign RE  =  Cntr[0];

always @(*)
begin
    case(Cntr[4:1])
        4'b0000 : DO <= 8'b00000001;
        4'b0001 : DO <= 8'b00000010;
        4'b0010 : DO <= 8'b00000100;
        4'b0011 : DO <= 8'b00001000;
        4'b0100 : DO <= 8'b00010000;
        4'b0101 : DO <= 8'b00100000;
        4'b0110 : DO <= 8'b01000000;
        4'b0111 : DO <= 8'b10000000;
        4'b1000 : DO <= 8'b01111111;
        4'b1001 : DO <= 8'b10111111;
        4'b1010 : DO <= 8'b11011111;
        4'b1011 : DO <= 8'b11101111;
        4'b1100 : DO <= 8'b11110111;
        4'b1101 : DO <= 8'b11111011;
        4'b1110 : DO <= 8'b11111101;
        4'b1111 : DO <= 8'b11111110;
    endcase
end

always @(posedge Clk)
begin
    if(Rst)
        rDO <= #1 0;
    else if(Cntr[0])
        rDO <= #1 DO;
end

always @(posedge Clk)
begin
    if(Rst)
        Err <= #1 0;
    else if(~Cntr[0])
        Err <= #1 Err | (rDO != DI);
end

assign LED = {Cntr[19:17], Cntr[5], ~Err};

ODDR2   #(
            .DDR_ALIGNMENT("NONE"), // Sets output alignment: "NONE", "C0", "C1"
            .INIT(1'b1),            // Sets initial state of the Q output: 1, 0
            .SRTYPE("ASYNC")        // Specifies "SYNC" or "ASYNC" set/reset
        ) PHI2 (
            .Q(Phi2O),              // 1-bit DDR output data
            .C0( Clk),              // 1-bit clock input
            .C1(~Clk),              // 1-bit clock input
            .CE(1'b1),              // 1-bit clock enable input
            .D0(1'b0),              // 1-bit data input (associated with C0)
            .D1(1'b1),              // 1-bit data input (associated with C1)
            .R(1'b0),               // 1-bit reset input
            .S(Rst)                 // 1-bit set input
        );

ODDR2   #(
            .DDR_ALIGNMENT("NONE"), // Sets output alignment: "NONE", "C0", "C1"
            .INIT(1'b1),            // Sets initial state of the Q output: 1, 0
            .SRTYPE("ASYNC")        // Specifies "SYNC" or "ASYNC" set/reset
        ) PHI1 (
            .Q(Phi1O),              // 1-bit DDR output data
            .C0(~Clk),              // 1-bit clock input
            .C1( Clk),              // 1-bit clock input
            .CE(1'b1),              // 1-bit clock enable input
            .D0(1'b0),              // 1-bit data input (associated with C0)
            .D1(1'b1),              // 1-bit data input (associated with C1)
            .R(1'b0),               // 1-bit reset input
            .S(Rst)                 // 1-bit set input
        );

endmodule
